/*
 * Copyright (c) 1996-2004 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as "Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

#include <stdio.h>
#include <ctype.h>
#include <string>

#define BPATCH_FILE

#include "process.h"
#include "debug.h"
#include "BPatch.h"
#include "BPatch_module.h"
#include "BPatch_libInfo.h"
#include "BPatch_function.h"
#include "BPatch_statement.h"
#include "BPatch_collections.h"
#include "symtabAPI/h/Type.h"    // For BPatch_type related stuff

#include "mapped_module.h"
#include "mapped_object.h"
#include "instPoint.h"

#if defined(arch_ia64)
#include "arch-ia64.h"
#endif

std::string current_func_name;
std::string current_mangled_func_name;
BPatch_function *current_func = NULL;

/*
 * BPatch_module::getSourceObj()
 *
 * Return the contained source objects (e.g. functions).
 *
 */

bool BPatch_module::getSourceObj(BPatch_Vector<BPatch_sourceObj *> &vect)
{
   if (!mod) return false;

   BPatch_Vector<BPatch_function *> temp;
   bool result = getProcedures(temp);
   if (!result)
      return false;

   for (unsigned int i = 0; i < temp.size(); ++i) {
      vect.push_back((BPatch_sourceObj*) temp[i]);
   }
   return true;
}

/*
 * BPatch_function::getObjParent()
 *
 * Return the parent of the function (i.e. the image)
 *
 */
BPatch_sourceObj *BPatch_module::getObjParent()
{
   return (BPatch_sourceObj *) img;
}

char *BPatch_module::getNameInt(char *buffer, int length)
{
   if (!mod)
      return NULL;

   string str = mod->fileName();

   strncpy(buffer, str.c_str(), length);

   return buffer;
}

const char *BPatch_module::libraryNameInt()
{
   if (!mod)
      return NULL;

   if (isSharedLib())      
      return mod->fullName().c_str();

   return NULL;
}

char *BPatch_module::getFullNameInt(char *buffer, int length)
{
   if (!mod)
      return NULL;
   string str = mod->fullName();

   strncpy(buffer, str.c_str(), length);

   return buffer;
}


BPatch_module::BPatch_module(BPatch_addressSpace *_addSpace, 
                             AddressSpace *as,
                             mapped_module *_mod,
                             BPatch_image *_img ) :
   addSpace(_addSpace), 
   lladdSpace(as),
   mod(_mod),
   img(_img), 
   moduleTypes(NULL),
   full_func_parse(false),
   full_var_parse(false)
{
   _srcType = BPatch_sourceModule;

   switch(mod->language()) {
      case lang_C:
         setLanguage( BPatch_c );
         break;

      case lang_CPlusPlus:
      case lang_GnuCPlusPlus:
         setLanguage( BPatch_cPlusPlus );
         break;

      case lang_Fortran_with_pretty_debug:
         setLanguage( BPatch_f90_demangled_stabstr );
         break;

      case lang_Fortran:
      case lang_CMFortran:
         setLanguage( BPatch_fortran );
         break;

      case lang_Assembly:
         setLanguage( BPatch_assembly );
         break;

      case lang_Unknown: 
      default:
         setLanguage( BPatch_unknownLanguage );
         break;
   }; /* end language switch */

} /* end BPatch_module() */

// Public 'destructor' function...
void BPatch_module::handleUnload() 
{
   // Hrm... what to do. For now, mark us as "deleted" so that
   // any other calls return an error.

   // Brainstorm: we can set mod to NULL (as it really is, 
   // having been deleted...) and key off that. Saves a boolean.

   mod = NULL;
}

bool BPatch_module::isValidInt() 
{
   return mod != NULL;
}

BPatch_module::~BPatch_module()
{
   if (moduleTypes) {
      BPatch_typeCollection::freeTypeCollection(moduleTypes);
   }

   // XXX The odd while-loop structures allow the
   //     destructors for map objects to remove themselves
   //     from the maps; otherwise the iterators will
   //     become invalidated

   BPatch_funcMap::iterator fi = func_map.begin();
   BPatch_funcMap::iterator tmp_f;
   while(fi != func_map.end()) {
      tmp_f = fi;
      ++fi;
      delete (*tmp_f).second;
   }

   BPatch_instpMap::iterator ii = instp_map.begin();
   BPatch_instpMap::iterator tmp_i;
   while(ii != instp_map.end()) {
      tmp_i = ii;
      ++ii; 
      delete (*tmp_i).second;
   }

   BPatch_varMap::iterator vi = var_map.begin();
   BPatch_varMap::iterator tmp_v;
   while(vi != var_map.end()) {
      tmp_v = vi;
      ++vi;
      delete (*tmp_v).second;
   }

   func_map.clear();
   instp_map.clear();
   var_map.clear();
}

bool BPatch_module::parseTypesIfNecessary() 
{
	if ( moduleTypes != NULL ) 
		return false;

	if (!isValid())
		return false;

	bool is64 = (mod->pmod()->imExec()->getAddressWidth() == 8);

	if (sizeof(void *) == 8 && !is64) 
	{
		// Terrible Hack:
		//   If mutatee and mutator address width is different,
		//   we need to patch up certain standard types.
		BPatch_type *typePtr;

		typePtr = BPatch::bpatch->builtInTypes->findBuiltInType(-10);
		typePtr->getSymtabType()->setSize(4);

		typePtr = BPatch::bpatch->builtInTypes->findBuiltInType(-19);
		typePtr->getSymtabType()->setSize(4);
	}

	mod->pmod()->mod()->exec()->parseTypesNow();
	moduleTypes = BPatch_typeCollection::getModTypeCollection(this);

	vector<Type *> *modtypes = mod->pmod()->mod()->getAllTypes();

	if (!modtypes)
		return false;

	for (unsigned i=0; i<modtypes->size(); i++) 
	{
		Type *typ = (*modtypes)[i];
		BPatch_type *type = new BPatch_type(typ);
		moduleTypes->addType(type);
	}

	vector<pair<string, Type *> > *globalVars = mod->pmod()->mod()->getAllGlobalVars();

	if (!globalVars)
		return false;

	for (unsigned i=0; i<globalVars->size(); i++)
	{
		BPatch_type *var_type = NULL;
		extern AnnotationClass<BPatch_type> TypeUpPtrAnno;

		Type *ll_var_type = (*globalVars)[i].second;
		std::string &var_name = (*globalVars)[i].first;

		assert(ll_var_type);

		if (!ll_var_type->getAnnotation(var_type, TypeUpPtrAnno))
		{
			var_type = new BPatch_type((*globalVars)[i].second);
		}
		else
		{
			assert(var_type);
		}

		moduleTypes->addGlobalVariable(var_name.c_str(), var_type);
#if 0
		if (!(*globalVars)[i].second->getUpPtr())
		{
			new BPatch_type((*globalVars)[i].second);
		}

		moduleTypes->addGlobalVariable((*globalVars)[i].first.c_str(), 
				(BPatch_type *)(*globalVars)[i].second->getUpPtr());
#endif
	}
	return true; 
}

BPatch_typeCollection *BPatch_module::getModuleTypesInt() 
{
	parseTypesIfNecessary();
	return moduleTypes;
}

/*
 * BPatch_module::getProcedures
 *
 * Returns a list of all procedures in the module upon success, and NULL
 * upon failure.
 */
BPatch_Vector<BPatch_function *> *
	BPatch_module::getProceduresInt(bool incUninstrumentable) {
		if (!isValid())
			return NULL;

   BPatch_Vector<BPatch_function*> *funcs = new BPatch_Vector<BPatch_function*>();
   bool result = getProceduresInt(*funcs, incUninstrumentable);
   if (!result) {
      delete funcs;
      return NULL;
   }

   return funcs;
}

bool BPatch_module::getProceduresInt(BPatch_Vector<BPatch_function*> &funcs,
                                     bool incUninstrumentable)
{
   if (!isValid())
      return false;

   if (!full_func_parse) {
      const pdvector<int_function*> &funcs = mod->getAllFunctions();
      for (unsigned i=0; i<funcs.size(); i++) {
         if (!func_map.count(funcs[i])) {
            addSpace->findOrCreateBPFunc(funcs[i], this);
         }
      }
      full_func_parse = true;
   }      

   BPatch_funcMap::iterator i = func_map.begin();
   for (; i != func_map.end(); i++) {
      if (incUninstrumentable || (*i).first->isInstrumentable())
         funcs.push_back((*i).second);
   }
   return true;
}

/*
 * BPatch_module::findFunction
 *
 * Returns a vector of BPatch_function* with the same name that is provided or
 * NULL if no function with that name is in the module.  This function
 * searches the BPatch_function vector of the module followed by
 * the int_function of the module.  If a int_function is found
 * a BPatch_function is created and added to the BPatch_function vector of
 * the module.
 * name The name of function to look up.
 */

   BPatch_Vector<BPatch_function *> *
BPatch_module::findFunctionInt(const char *name, 
      BPatch_Vector<BPatch_function *> & funcs,
      bool notify_on_failure, bool regex_case_sensitive,
      bool incUninstrumentable, bool dont_use_regex)
{
   if (!isValid())
      return NULL;

   unsigned size = funcs.size();

   if (!name) {
      char msg[512];
      sprintf(msg, "%s[%d]:  Module %s: findFunction(NULL)...  failing",
            __FILE__, __LINE__, mod->fileName().c_str());
      BPatch_reportError(BPatchSerious, 100, msg);
      return NULL;
   }

   // Do we want regex?
   if (dont_use_regex 
         ||  (NULL == strpbrk(name, REGEX_CHARSET))) {
      pdvector<int_function *> int_funcs;
      if (mod->findFuncVectorByPretty(name, int_funcs)) {
         for (unsigned piter = 0; piter < int_funcs.size(); piter++) {
            if (incUninstrumentable || int_funcs[piter]->isInstrumentable()) 
            {
               BPatch_function * bpfunc = addSpace->findOrCreateBPFunc(int_funcs[piter], this);
               funcs.push_back(bpfunc);
            }
         }
      }
      else {
         if (mod->findFuncVectorByMangled(name,
                  int_funcs)) {
            for (unsigned miter = 0; miter < int_funcs.size(); miter++) {
               if (incUninstrumentable || int_funcs[miter]->isInstrumentable()) 
               {
                  BPatch_function * bpfunc = addSpace->findOrCreateBPFunc(int_funcs[miter], this);
                  //						BPatch_function * bpfunc = proc->findOrCreateBPFunc(int_funcs[miter], this);
                  funcs.push_back(bpfunc);
               }
            }
         }
      }
      if (size != funcs.size())
         return &funcs;
   }
   else {
      // Regular expression search. As with BPatch_image, we handle it here

#if !defined(os_windows)
      // REGEX falls through:
      regex_t comp_pat;
      int err, cflags = REG_NOSUB | REG_EXTENDED;

      if ( !regex_case_sensitive )
         cflags |= REG_ICASE;

      //cerr << "compiling regex: " <<name<<endl;

      if (0 != (err = regcomp( &comp_pat, name, cflags ))) {
         char errbuf[80];
         regerror( err, &comp_pat, errbuf, 80 );
         if (notify_on_failure) {
            cerr << __FILE__ << ":" << __LINE__ << ":  REGEXEC ERROR: "<< errbuf << endl;
            std::string msg = std::string("Image: Unable to find function pattern: ") 
               + std::string(name) + ": regex error --" + std::string(errbuf);
            BPatch_reportError(BPatchSerious, 100, msg.c_str());
         }
         return NULL;
      }

      // Regular expression search. This used to be handled at the image
      // class level, but was moved up here to simplify semantics. We
      // have to iterate over every function known to the process at some
      // point, so it might as well be top-level. This is also an
      // excellent candidate for a "value-added" library.

      const pdvector<int_function *> &int_funcs = mod->getAllFunctions();

      for (unsigned ai = 0; ai < int_funcs.size(); ai++) {
         int_function *func = int_funcs[ai];
         // If it matches, push onto the vector
         // Check all pretty names (and then all mangled names if there is no match)
         bool found_match = false;
         for (unsigned piter = 0; piter < func->prettyNameVector().size(); piter++) {
            const string &pName = func->prettyNameVector()[piter];
            int err;     
            if (0 == (err = regexec(&comp_pat, pName.c_str(), 1, NULL, 0 ))){
               if (func->isInstrumentable() || incUninstrumentable) {
                  BPatch_function *foo = addSpace->findOrCreateBPFunc(func, NULL);
                  //	   BPatch_function *foo = proc->findOrCreateBPFunc(func, NULL);
                  funcs.push_back(foo);
               }
               found_match = true;
               break;
            }
         }
         if (found_match) continue; // Don't check mangled names

         for (unsigned miter = 0; miter < func->symTabNameVector().size(); miter++) {
            const string &mName = func->symTabNameVector()[miter];
            int err;

            if (0 == (err = regexec(&comp_pat, mName.c_str(), 1, NULL, 0 ))){
               if (func->isInstrumentable() || incUninstrumentable) {
                  BPatch_function *foo = addSpace->findOrCreateBPFunc(func, NULL);
                  //	   BPatch_function *foo = proc->findOrCreateBPFunc(func, NULL);
                  funcs.push_back(foo);
               }
               found_match = true;
               break;
            }
         }
      }

      regfree(&comp_pat);

      if (funcs.size() != size) {
         return &funcs;
      } 

      if (notify_on_failure) {
         std::string msg = std::string("Unable to find pattern: ") + std::string(name);
         BPatch_reportError(BPatchSerious, 100, msg.c_str());
      }
#endif
   }

   if (notify_on_failure) {
      char msg[1024];
      sprintf(msg, "%s[%d]:  Module %s: unable to find function %s",
            __FILE__, __LINE__, mod->fileName().c_str(), name);
      BPatch_reportError(BPatchSerious, 100, msg);

   }
   return &funcs;
}

   BPatch_Vector<BPatch_function *> *
BPatch_module::findFunctionByAddressInt(void *addr, BPatch_Vector<BPatch_function *> &funcs,
      bool notify_on_failure, 
      bool incUninstrumentable)
{
   if (!isValid()) return NULL;

   int_function *pdfunc = NULL;
   BPatch_function *bpfunc = NULL;

   pdfunc = mod->findFuncByAddr((Address)addr);
   if (!pdfunc) {
      if (notify_on_failure) {
         char msg[1024];
         sprintf(msg, "%s[%d]:  Module %s: unable to find function %p",
               __FILE__, __LINE__, mod->fileName().c_str(), addr);
         BPatch_reportError(BPatchSerious, 100, msg);
      }
      return NULL;
   }

   if (incUninstrumentable || pdfunc->isInstrumentable()) {
      bpfunc = addSpace->findOrCreateBPFunc(pdfunc, this);
      //bpfunc = proc->findOrCreateBPFunc(pdfunc, this);
      if (bpfunc) {
         funcs.push_back(bpfunc);
      }
   }
   return &funcs;
}

BPatch_function * BPatch_module::findFunctionByMangledInt(const char *mangled_name,
      bool incUninstrumentable)
{
   if (!isValid()) return NULL;

   BPatch_function *bpfunc = NULL;

   pdvector<int_function *> int_funcs;
   std::string mangled_str(mangled_name);

   if (!mod->findFuncVectorByMangled(mangled_str,
            int_funcs))
      return NULL;

   if (int_funcs.size() > 1) {
      fprintf(stderr, "%s[%d]: Warning: found multiple name matches for %s, returning first\n",
            FILE__, __LINE__, mangled_name);
   }

   int_function *pdfunc = int_funcs[0];

   if (incUninstrumentable || pdfunc->isInstrumentable()) {
      bpfunc = addSpace->findOrCreateBPFunc(pdfunc, this);
   }

   return bpfunc;
}

bool BPatch_module::dumpMangledInt(char * prefix)
{
   mod->dumpMangled(prefix);
   return true;
}

void BPatch_module::parseTypes() 
{
   mod->pmod()->mod()->exec()->parseTypesNow();
}
// This is done by analogy with BPatch_module::getVariables,
// not BPatch_image::findVariable.  This should result in consistent
// behavior at the module level.
BPatch_variableExpr* BPatch_module::findVariableInt(const char* name)
{
   parseTypesIfNecessary();
   const pdvector<int_variable *> &allVars = mod->getAllVariables();

   for (unsigned i = 0; i < allVars.size(); i++) {
     if(strcmp(allVars[i]->symTabName().c_str(), name) == 0)
     {
        return addSpace->findOrCreateVariable(allVars[i]);
     }
   }

   return NULL;
}

bool BPatch_module::getVariablesInt(BPatch_Vector<BPatch_variableExpr *> &vars)
{
   if (!isValid())
      return false;
   if (!full_var_parse) {
      const pdvector<int_variable*> &vars = mod->getAllVariables();
      for (unsigned i=0; i<vars.size(); i++) {
         if (!var_map.count(vars[i])) {
            addSpace->findOrCreateVariable(vars[i]);
         }
      }
      full_var_parse = true;
   }      

   BPatch_varMap::iterator i = var_map.begin();
   for (; i != var_map.end(); i++) {
      vars.push_back((*i).second);
   }

   return true;
}

/* This function should be deprecated. */
bool BPatch_module::getLineToAddrInt( unsigned int lineNo, 
      BPatch_Vector< unsigned long > & buffer, bool ) 
{
   if (!isValid()) {
      fprintf(stderr, "%s[%d]: module is not valid\n", FILE__, __LINE__);
      return false;
   }

   std::vector< std::pair< Address, Address > > ranges;
   if ( ! getAddressRangesInt( NULL, lineNo, ranges ) ) { 
      return false; 
   }

   for ( unsigned int i = 0; i < ranges.size(); ++i ) {
      buffer.push_back( ranges[i].first + mod->obj()->codeBase());
   }

   return true;
} /* end getLineToAddr() */

bool BPatch_module::getSourceLinesInt(unsigned long addr, 
      BPatch_Vector< BPatch_statement> &lines) 
{
   if (!isValid()) 
   {
      fprintf(stderr, "%s[%d]:  failed to getSourceLines: invalid\n", FILE__, __LINE__);
      return false;
   }

   unsigned int originalSize = lines.size();
   LineInformation *li = mod->getLineInformation();
   std::vector<LineNoTuple> lines_ll;

   if (li) 
   {
      li->getSourceLines( addr - mod->obj()->codeBase(), lines_ll );
   }
   else 
   {
      return false;
   }

   for (unsigned int j = 0; j < lines_ll.size(); ++j) 
   {
      LineNoTuple &t = lines_ll[j];
      lines.push_back(BPatch_statement(this, t.first, t.second, t.column));
   }

   return (lines.size() != originalSize);
} /* end getSourceLines() */

bool BPatch_module::getStatementsInt(BPatch_Vector<BPatch_statement> &statements)
{
   // Iterate over each address range in the line information
   LineInformation *li = mod->getLineInformation();
   if (!li) {
      return false;
   }

   for (LineInformation::const_iterator i = li->begin();
         i != li->end();
         ++i) {

      // Form a BPatch_statement object for this entry
      // Note:  Line information stores offsets, so we need to adjust to
      //  addresses
      BPatch_statement statement(this, i->second.first, i->second.second,
            i->second.column, 
            (void *)(i->first.first + mod->obj()->codeBase()), 
            (void *)(i->first.second + mod->obj()->codeBase()));

      // Add this statement
      statements.push_back(statement);

   }
   return true;
}

bool BPatch_module::getAddressRangesInt( const char * fileName, 
      unsigned int lineNo, std::vector< std::pair< Address, Address > > & ranges ) 
{
   unsigned int starting_size = ranges.size();

   if (!isValid()) 
   {
      fprintf(stderr, "%s[%d]:  module is not valid\n", FILE__, __LINE__);
      return false;
   }

   LineInformation *li = mod->getLineInformation();

   if ( fileName == NULL ) 
   {
      fileName = mod->fileName().c_str(); 
   }

   if (li) 
   {
      bool ok = li->getAddressRanges( fileName, lineNo, ranges );
      if (!ok)
         return false;

      //  Iterate over the returned offset ranges to turn them into addresses
      for (unsigned int i = starting_size; i < ranges.size(); ++i) 
      {
         ranges[i].first += mod->obj()->codeBase();
         ranges[i].second += mod->obj()->codeBase();
      }
      
      return true;
   }

   return false;
} /* end getAddressRanges() */

bool BPatch_module::isSharedLibInt() 
{
   return mod->obj()->isSharedLib();
}

/*
 * BPatch_module::getBaseAddr
 *
 * Returns the starting address of the module.
 */
void *BPatch_module::getBaseAddrInt()
{
   return (void *)mod->obj()->codeAbs();
}

/*
 * BPatch_module::getSize
 *
 * Returns the size of the module in bytes.
 */
unsigned long BPatch_module::getSizeInt() 
{
   if (!mod) return 0;
   return (unsigned long) mod->obj()->imageSize();
}


bool BPatch_module::isNativeCompilerInt()
{
   if (!mod) return false;
   return mod->obj()->parse_img()->isNativeCompiler();
}

size_t BPatch_module::getAddressWidthInt()
{
   if (!mod) return 0;
   return mod->obj()->parse_img()->getObject()->getAddressWidth();
}

void BPatch_module::setDefaultNamespacePrefix(char * /*name*/) 
{ 
}

bool BPatch_module::isSystemLib() 
{
   if (!mod) return false;
   string str = mod->fileName();

   // Solaris 2.8... we don't grab the initial func always,
   // so fix up this code as well...
#if defined(os_solaris)
   if (strstr(str.c_str(), "libthread"))
      return true;
#endif
#if defined(os_linux)
   if (strstr(str.c_str(), "libc.so"))
      return true;
   if (strstr(str.c_str(), "libpthread"))
      return true;
#endif

   if (strstr(str.c_str(), "libdyninstAPI_RT"))
      return true;
#if defined(os_windows)
   if (strstr(str.c_str(), "kernel32.dll"))
      return true;
   if (strstr(str.c_str(), "user32.dll"))
      return true;
   if (strstr(str.c_str(), "ntdll.dll"))
      return true;
#endif

   return false;
}

AddressSpace *BPatch_module::getAS()
{
   return lladdSpace;
}

/* Build up a vector of control flow instructions with suspicious targets
 * 1. Gather all instructions with static targets that leave the low-level
 *    mapped_object's image and check to see if they jump into another
 *    existing module
 * 2. Add all dynamic call instructions to the vector
 * 3. Add all dynamic jump instructions to the vector
 */
BPatch_Vector<BPatch_point *> *BPatch_module::getUnresolvedControlFlowInt()
{
    if (unresolvedCF.size()) {
        unresolvedCF.clear();
    }
    std::set<image_instPoint*> ctrlTransfers 
        = lowlevel_mod()->pmod()->getUnresolvedControlFlow();

    std::set<image_instPoint*>::iterator cIter = ctrlTransfers.begin();
    while (cIter != ctrlTransfers.end()) {
        image_instPoint *curPoint = *cIter;
        Address curAddr = curPoint->offset() 
            + curPoint->func()->img()->desc().loadAddr();
        // if this is a static CT skip if target leads to known code
        if ( ! curPoint->isDynamic() && 
             getAS()->findOrigByAddr(curPoint->callTarget())) {
            continue;
        }
        // find or create BPatch_point from image_instPoint
        BPatch_point *bpPoint = NULL;
        BPatch_function *func = img->findFunctionInt
            ((unsigned long)curAddr); // takes absolute address
        if (func) {
            // Create the vector of call points in the low-level func
            if (curPoint->getPointType() == callSite) {
                func->lowlevel_func()->funcCalls();
            }
            // takes absolute address
            instPoint *intPt = func->lowlevel_func()
                ->findInstPByAddr(curAddr);
            if (!intPt) {// instPoint does not exist, create it
                intPt = instPoint::createParsePoint
                    (func->lowlevel_func(), curPoint);
            }
            if (!intPt) {
                fprintf(stderr,"%s[%d] Did not find instPoint for "
                        "unresolved control transfer at %lx\n",
                        __FILE__,__LINE__, (long)curAddr);
                continue;
            }
            if (intPt->getPointType() == callSite) {
                bpPoint = addSpace->findOrCreateBPPoint
                    (func, intPt, BPatch_locSubroutine);
            } else {
                bpPoint = addSpace->findOrCreateBPPoint
                    (func, intPt, BPatch_locLongJump);
            }
        } else {
            fprintf(stderr,"%s[%d] Could not find function corresponding to "
                    "unresolved control transfer at %lx\n",
                    __FILE__,__LINE__, (long)curAddr);
        } 
        // add BPatch_point to vector of unresolved control transfers
        if (bpPoint == NULL) {
            fprintf(stderr,"%s[%d] Control transfer at %lx not instrumentable\n",
                    __FILE__,__LINE__, (long)curAddr);
        } else {
            unresolvedCF.push_back(bpPoint);
        }
        cIter++;
    }
    return &unresolvedCF;
}

#ifdef IBM_BPATCH_COMPAT

bool BPatch_module::getLineNumbersInt( unsigned int & startLine, unsigned int & endLine )
{
   /* I don't think this function has ever returned nonzeroes.  Approximate a better 
      result by with the line numbers for the first and last addresses in the module. */
   if (!mod) return false;

   void * startAddr, * endAddr;
   if( ! getAddressRangeInt( startAddr, endAddr ) ) {
      return false;
   }

   bool setAValue = false;
   BPatch_Vector<BPatch_statement> lines;
   getSourceLines( (Address)startAddr, lines );
   if( lines.size() != 0 ) {
      startLine = lines[0].lineNumber();
      setAValue = true;
   }

   lines.clear();
   getSourceLines( (Address)endAddr, lines );
   if( lines.size() != 0 ) {
      endLine = lines[0].lineNumber();
      setAValue = true;
   }

   return setAValue;
}

bool BPatch_module::getAddressRangeInt(void * &start, void * &end)
{
   // Code? Data? We'll do code for now...
   if (!mod) return false;
   start = (void *)(mod->obj()->codeAbs());
   end = (void *)(mod->obj()->codeAbs() + mod->obj()->imageSize());
   return true;
}
char *BPatch_module::getUniqueStringInt(char *buffer, int length)
{
   // Use "<program_name>|<module_name>" as the unique name if this module is
   // part of the executable and "<module_name>" if it is not.
   if (!mod) return NULL;
   if(isSharedLib())
      snprintf(buffer, length, "%s", mod->fileName().c_str());
   else {
      char prog[1024];
      addSpace->getImage()->getProgramFileName(prog, 1024);
      snprintf(buffer, length, "%s|%s",
            prog, mod->fileName().c_str());
   }
   // Return the unique name to the caller
   return buffer;
}

int BPatch_module::getSharedLibTypeInt()	
{
   return 0;
}

int BPatch_module::getBindingTypeInt()
{
   return 0;
}

std::vector<struct BPatch_module::Statement> BPatch_module::getStatementsInt()
{
   std::vector<struct BPatch_module::Statement> statements;
   if (!mod) return statements;

   LineInformation *li = mod->getLineInformation();
   if (!li)
      return statements;

   // Iterate over each address range in the line information
   for (LineInformation::const_iterator i = li->begin(); i != li->end(); ++i) {
      // Form a Statement object for this entry
      struct BPatch_module::Statement statement;
      statement.begin = i->first.first + mod->obj()->codeBase();
      statement.end = i->first.second + mod->obj()->codeBase();
      statement.path = i->second.first;
      statement.line = i->second.second;
      statement.column = 0;

      // Add this statement
      statements.push_back(statement);

   }

   // Return the statements to the caller
   return statements;
}
#endif
