/*
 * Copyright (c) 1996-2004 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as "Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

/*
 * inst-x86.C - x86 dependent functions and code generator
 */

#include "common/h/Vector.h"
#include "common/h/Dictionary.h"
#include "common/h/Vector.h"
#include "image-func.h"
#include "instPoint.h"
#include "symtab.h"
#include "dyninstAPI/h/BPatch_Set.h"
#include "InstrucIter.h"
#include "debug.h"
#include <deque>
#include <set>
#include <algorithm>
#include "arch.h"
#if defined(cap_instruction_api)
#include "instructionAPI/h/Instruction.h"
#include "instructionAPI/h/InstructionDecoder.h"
#else
namespace Dyninst
{
	namespace InstructionAPI
	{
		struct RegisterAST
		{
			typedef boost::shared_ptr<RegisterAST> Ptr;
		};
	};
};
#endif
/**************************************************************
 *
 *  machine dependent methods of pdFunction
 *
 **************************************************************/

void checkIfRelocatable(instruction insn, bool &canBeRelocated) {
  const unsigned char *instr = insn.ptr();

  // Check if REG bits of ModR/M byte are 100 or 101 (Possible jump 
  // to jump table).
  if (instr[0] == 0xFF && 
     ( ((instr[1] & 0x38)>>3) == 4 || ((instr[1] & 0x38)>>3) == 5 )) {

    // function should not be relocated
    canBeRelocated = false;
  }
}

bool image_func::archIsIPRelativeBranch(InstrucIter& ah)
{
	// These don't exist on IA32...
#if !defined(arch_x86_64)
  return false;
#endif
  instruction branchCandidate = ah.getInstruction();
  // Branch candidate should have jump opcode, mod r/m, displacement...
  if(branchCandidate.isJumpIndir())
  {
    ia32_locations locs;
    ia32_memacc mac[3];
    ia32_condition c;
    ia32_instruction ll_insn(mac, &c, &locs);
    ia32_decode(IA32_FULL_DECODER,
		static_cast<const unsigned char*>(img()->getPtrToInstruction(*ah)),
		ll_insn);
    if(locs.modrm_mod == 0x0 && locs.modrm_rm == 0x05)
    {
      parsing_printf("%s[%d]: Found RIP-offset indirect branch at 0x%lx\n", FILE__,
		     __LINE__, *ah);
      return true;
    }
  }
  return false;
}

bool image_func::archIsRealCall(InstrucIter &ah, bool &validTarget,
                                bool & /* simulateJump */)
{
    instruction insn = ah.getInstruction();
    Address adr = *ah;

   // Initialize return value
   validTarget = true;
   //parsing_printf("*** Analyzing call, offset 0x%x target 0x%x\n",
   //adr, insn.getTarget(adr));
   // calls to adr+5 are not really calls, they are used in 
   // dynamically linked libraries to get the address of the code.
   if (ah.isADynamicCallInstruction()) {
     parsing_printf("... Call 0x%lx is indirect\n", adr);
     validTarget = false;
     return true;
   }
   if (insn.getTarget(adr) == adr + 5) {
       parsing_printf("... getting PC\n");
       // XXX we can do this like on sparc, but we don't need to: we do it
       // on sparc because the intervening instructions don't get executed;
       // on x86 for this heuristic there *are* no intervening instructions.
       //simulateJump = true;
       return false;
   }

   // Calls to a mov instruction followed by a ret instruction, where the 
   // source of the mov is the %esp register, are not real calls.
   // These sequences are used to set the destination register of the mov 
   // with the pc of the instruction instruction that follows the call.

   // This sequence accomplishes this because the call instruction has the 
   // side effect of placing the value of the %eip on the stack and setting the
   // %esp register to point to that location on the stack. (The %eip register
   // maintains the address of the next instruction to be executed).
   // Thus, when the value at the location pointed to by the %esp register 
   // is moved, the destination of the mov is set with the pc of the next
   // instruction after the call.   

   //    Here is an example of this sequence:
   //
   //       mov    %esp, %ebx
   //       ret    
   //
   //    These two instructions are specified by the bytes 0xc3241c8b
   //

   Address targetOffset = insn.getTarget(adr);
 
   if ( !img()->isValidAddress(targetOffset) ) {
       parsing_printf("... Call to 0x%lx is invalid (outside code or data)\n",
       targetOffset);
       validTarget = false;
       return false;
   }    

   // Get a pointer to the call target
   const unsigned char *target =
      (const unsigned char *)img()->getPtrToInstruction(targetOffset);

   // The target instruction is a  mov
   if (*(target) == 0x8b) {
      // The source register of the mov is specified by a SIB byte 
      if (*(target + 1) == 0x1c || *(target + 1) == 0x0c) {

         // The source register of the mov is the %esp register (0x24) and 
         // the instruction after the mov is a ret instruction (0xc3)
         if ( (*(target + 2) == 0x24) && (*(target + 3) == 0xc3)) {
            return false;
         }
      }
   }

   return true;
}

// Determine if the called function is a "library" function or a "user"
// function This cannot be done until all of the functions have been seen,
// verified, and classified
// 
// DELAYED UNTIL PROCESS SPECIALIZATION
//






/********************************************************/
/* Architecture dependent parsing support methods       */
/********************************************************/

bool image_func::archCheckEntry( InstrucIter &ah, image_func * /*func*/ )
{
    instruction insn = ah.getInstruction();

  // check if the entry point contains another point
  if (insn.isJumpDir()) 
    {
      
      return false;
    } 
  else if (insn.isReturn()) 
    {
      // this is an empty function
      return false;
    } 
  else if (insn.isCall()) 
    {
        // So?
        return true;
    }
  return true;
}

// Architecture-specific a-priori determination that we can't
// parse this function.
bool image_func::archIsUnparseable()
{
    if( !isInstrumentableByFunctionName() )
    {   
        if (!isInstrumentableByFunctionName())
            parsing_printf("... uninstrumentable by func name\n");

        //endOffset_ = startOffset_;
        endOffset_ = getOffset();
        instLevel_ = UNINSTRUMENTABLE; 
        return true;
    }           
    else
        return false;
}

// Architecture-specific hack to give up happily on parsing a
// function.
bool image_func::archAvoidParsing()
{
    //temporary convenience hack.. we don't want to parse the PLT as a function
    //but we need pltMain to show up as a function
    //so we set size to zero and make sure it has no instPoints.    
    if( prettyName() == "DYNINST_pltMain" )//|| 
        //prettyName() == "winStart" ||
        //prettyName() == "winFini" )
    {   
        //endOffset_ = startOffset_;
        endOffset_ = getOffset();
        return true;
    }
    else
        return false;
}

void image_func::archGetFuncEntryAddr(Address & /* funcEntryAddr */)
{
    return;
}

// Architecture-specific hack to prevent relocation of certain functions.
bool image_func::archNoRelocate()
{   
    return prettyName() == "__libc_start_main";
}

// Nop on x86
void image_func::archSetFrameSize(int /* frameSize */)
{
    return;
}

void image_func::archInstructionProc(InstrucIter & /* ah */)
{
    return;
}

bool findMaxSwitchInsn(image_basicBlock *start, instruction &maxSwitch,
                       instruction &branchInsn, bool &found_along_taken_br,
                       const std::set<Dyninst::InstructionAPI::RegisterAST::Ptr>& /*regsRead*/)
{
   using namespace Dyninst::InstructionAPI;

    BPatch_Set<image_basicBlock *> visited;
    pdvector<image_basicBlock *> WL;
    pdvector<image_edge *> sources;
   pdvector<image_edge *> targets;
    image_basicBlock *curBlk;
    int depth = 0;

    bool foundMaxSwitch = false;
    bool foundCondBranch = false;
    Address maxSwitchAddr = 0;

    WL.push_back(start);

    for(unsigned j=0;j < WL.size(); j++)
    {
        curBlk = WL[j];
        visited.insert(curBlk);

        foundMaxSwitch = false;
        foundCondBranch = false;
    
        InstrucIter iter( curBlk );
        while( *iter < curBlk->endOffset() ) {
            // check for cmp 
            if( iter.getInstruction().isCmp() )
            {
#if defined(unsafe_safety_check)
                // This is the right thing to do, but we see several examples of
                // code like so:

                //
                // cmp 0x13, 0x8(%ebp)
                // ... jump around
                // mov 0x8(%ebp), %eax
                // ... calculate
                // jmp *%eax
                // 
                // Note that there's an alias problem here that means we can't look for
                // cmp eax, ...

                InstructionDecoder d((const unsigned char *)curBlk->getFirstFunc()->img()->getPtrToInstruction(*iter), 
                                     iter.getInstruction().size());
                Instruction cmpInsn = d.decode();
                for(std::set<RegisterAST::Ptr>::const_iterator curReg = regsRead.begin();
                    curReg != regsRead.end();
                    ++curReg)
                    {
                        if(cmpInsn.isRead(*curReg))
                            {
                                parsing_printf("\tFound jmp table cmp instruction at 0x%lx\n",
                                               *iter);
                                maxSwitch = iter.getInstruction();
                                maxSwitchAddr = *iter;
                                foundMaxSwitch = true;
                            }
                    }
#else
            parsing_printf("\tFound jmp table cmp instruction at 0x%lx\n",
                           *iter);
                maxSwitch = iter.getInstruction();
                maxSwitchAddr = *iter;
                foundMaxSwitch = true;
#endif
            }
         if (iter.getInstruction().type() & IS_JCC) {
                parsing_printf("\tFound jmp table cond br instruction at 0x%lx\n",
                    *iter);
                branchInsn = iter.getInstruction();
                foundCondBranch = true;

            targets.clear();
            curBlk->getTargets(targets);
            bool taken_hit = false;
            bool fallthrough_hit = false;
            for (unsigned i=0; i<targets.size(); i++) {
               if (targets[i]->getType() == ET_COND_TAKEN &&
                   visited.contains(targets[i]->getTarget()))
               {
                  taken_hit = true;
               }
               if ((targets[i]->getType() == ET_COND_NOT_TAKEN ||
                    targets[i]->getType() == ET_FALLTHROUGH) &&
                   visited.contains(targets[i]->getTarget()))
               {
                  fallthrough_hit = true;
               }
            }
	    parsing_printf("\tfindMaxSwitchInsn: taken_hit: %d, fallthrough_hit: %d\n", taken_hit, fallthrough_hit);
            found_along_taken_br = taken_hit && !fallthrough_hit;
                break;
            }
            iter++;
        }

      if(foundMaxSwitch && foundCondBranch)
         break; // done

            // look further back
            sources.clear();
            curBlk->getSources( sources );
	    depth++;
	    // We've seen depth 2 in libc et al
	    if(depth > 2) return false;
	    
            for(unsigned i=0;i<sources.size();i++)
            {
                if(sources[i]->getType() == ET_CALL)
                    return false;

                image_basicBlock * src = sources[i]->getSource();
                if( !visited.contains( src ) ) {
                    WL.push_back(src);
                }
            }
        }
#if defined (cap_use_pdvector)
    WL.zap();
#else
    WL.clear();
#endif
    parsing_printf("\tfindMaxSwitchInsn: table on taken branch: %d, returning: %d\n", found_along_taken_br, foundMaxSwitch && foundCondBranch);
    
    return foundMaxSwitch && foundCondBranch;
}

bool isNonCallEdge(image_edge* e)
{
  return e->getType() != ET_CALL;
}
bool leadsToVisitedBlock(image_edge* e, const std::set<image_basicBlock*>& visited)
{
  image_basicBlock* src = e->getSource();
  return visited.find(src) != visited.end();
}

bool findThunkInBlock(image_func* f, image_basicBlock* curBlock, Address& thunkOffset)
{
  InstrucIter ah(curBlock);
  bool isValidTarget, dummy;
  
  while(ah.hasMore())
  {
    if(ah.isACallInstruction() && !f->archIsRealCall(ah, isValidTarget, dummy))
    {
      ah++;
      thunkOffset = *ah;
      ia32_memacc mac[3];
      ia32_locations loc;
      ia32_condition cond;
      ia32_instruction itmp(mac, &cond, &loc);
      ia32_decode(IA32_FULL_DECODER, const_cast<unsigned char*>(ah.getInstruction().ptr()), itmp);
      int immediate;
      if(itmp.getEntry()->id == e_add)
      {
	switch(loc.imm_size)
	{
	case 4:
	  immediate = *(const int*)(ah.getInstruction().ptr() + loc.imm_position);
      parsing_printf("\tsetting thunkOffset to 0x%lx (0x%lx + 0x%lx) (1)\n",thunkOffset+immediate,thunkOffset,immediate);
	  thunkOffset += immediate;
      return true;
	  break;
	case 1:
	  immediate = *(const char*)(ah.getInstruction().ptr() + loc.imm_position);
      parsing_printf("\tsetting thunkOffset to 0x%lx (0x%lx + 0x%lx) (2)\n",thunkOffset+immediate,thunkOffset,immediate);
	  thunkOffset += immediate;
      return true;
	  break;
	default:
	  thunkOffset = 0;
	  break;
	}
      }
    }
    else // Look for an AMD64 IP-relative LEA.  If we find one, it should point to the start of a
    {    // relative jump table.

      ia32_memacc mac[3];
      ia32_locations loc;
      ia32_condition cond;
      ia32_instruction itmp(mac, &cond, &loc);
      ia32_decode(IA32_FULL_DECODER, const_cast<unsigned char*>(ah.getInstruction().ptr()), itmp);

      if(itmp.getEntry()->id == e_lea && 
	 loc.modrm_rm == 0x05 && loc.modrm_mod == 0x00) // IP-relative
      {
	int immediate;
	switch(loc.disp_size)
	{
	case 4:
	  immediate = *(const int*)(ah.getInstruction().ptr() + loc.disp_position);
	  ah++;
	  thunkOffset = *ah + immediate;
      parsing_printf("\tsetting thunkOffset to 0x%lx (0x%lx + 0x%lx) (3)\n",thunkOffset,*ah,immediate);
      return true;
	  break;
	case 1:
	  immediate = *(const char*)(ah.getInstruction().ptr() + loc.disp_position);
	  ah++;
	  thunkOffset = *ah + immediate;
      parsing_printf("\tsetting thunkOffset to 0x%lx (0x%lx + 0x%lx) (4)\n",thunkOffset,*ah,immediate);
      return true;
	  break;
	default:
	  thunkOffset = 0;
	  break;	  
	  }
      }
    }
    if(ah.hasMore()) {
      ah++;
    }
  }
  return false;
}


bool findThunkAndOffset(image_func* f, image_basicBlock* start, Address& thunkOffset)
{
  std::set<image_basicBlock*> visited;
  std::deque<image_basicBlock*> worklist;
  pdvector<image_edge*> sources;
  image_basicBlock* curBlock;
  worklist.insert(worklist.begin(), start);
  visited.insert(start);
  while(!worklist.empty())
  {
    curBlock = worklist.front();
    worklist.pop_front();
    if(findThunkInBlock(f, curBlock, thunkOffset))
    {
      return true;
    }

      sources.clear();
      curBlock->getSources(sources);
    for (unsigned i=0; i<sources.size(); i++) {
       image_edge *e = sources[i];
       if(isNonCallEdge(e) && !leadsToVisitedBlock(e, visited))
       {
          worklist.push_back(e->getSource());
          visited.insert(e->getSource());
       }
    }
  }
  return false;
}


/* Recognizes particular ``jump table'' idioms on IA32/x86-64 platforms.
 
   There are two pieces of information needed:

   1. The base address of the table. This can be computed in one of two
      ways:
      - As an immediate value encoded in the indirect branch instruction
      - As an offset from a thunk

   2. The size of the table (number of indices). Knowing this value
      requires knowing two more things:
      - The address size in length (4 or 8 bytes)
      - The stride length of the addressing method used to compute
        offsets into the table (4 or 8 bytes)

   Note:
    The code below is extremely ad-hoc and recognizes several different
    jump tables. Having developed in multiple iterations over a long
    period of time, it is currently /hard to read/, in particular using
    redundant information in multiple contexts. This routine and its
    callees may require /cleanup/ to meet Wikipedia's /quality standards/.

   Suggestions for future improvement:
    1. Segregate functionality for jump table parsing into different
       routines corresponding to the different idioms we can recognize.
    
    2. Consider register identities when searching for comparison and
       load instructions used for the indirect branch. Currently none
       of the routines take this into account.
       /This is a serious deficiency/
*/
bool image_func::archGetMultipleJumpTargets( 
                                BPatch_Set< Address >& targets,
                                image_basicBlock * currBlk,
                                InstrucIter &ah,
                                pdvector< instruction >& allInstructions)
{
    using namespace Dyninst::InstructionAPI;

    //we are going to get the instructions to parse the 
    //jump table from my source block(s)
    pdvector< image_edge* > in;
    currBlk->getSources( in );

   if (archIsIPRelativeBranch(ah)) {
      return false;
   }
     
    Address tableInsnAddr;
   Address jumpAddr = *ah;

    if( in.size() < 1 )
    {
      return false;
    }
        // Assume that the table base address is calculated in the
        // branch instruction (encoded using SIB)
        instruction tableInsn = ah.getInstruction();
        bool isAddInJmp = true;
   bool foundJccAlongTaken;
        tableInsnAddr = *ah;

        // branchInsn refers to the conditional branch that guards
        // the jump table, not the indirect branch.
        instruction branchInsn;
        instruction maxSwitch;

        // There must be at least 3 instructions to form a jump table
        // idiom. XXX we don't even use this on IA32/x86-64!
        assert((allInstructions.size() - 2) > 0);
        
        const unsigned char* ptr = ah.getInstruction().op_ptr();
        const unsigned char LEA = 0x8D;

        // Instruction must be in ``group 5'' (includes indirect call
        // and branch)	
        assert( *ptr == 0xff );

        // If rm portion of mod/rm byte is not 0x4, then this instruction
        // does not use SIB addressing and so we assume (correct?) that
        // it cannot encode the address. Search backward for /any move
        // from reg/memory -> reg OR an LEA instruction that populates a 
        // register.
        //  
        // FIXME this is the first place where /no correlation between the
        // target register of the branch is required/
        ptr++;
        if((*ptr & 0xc7) != 0x04) // if not SIB
        {
            isAddInJmp = false;
            bool foundTableInsn = false;
                
            InstrucIter findReg(ah);

            while(findReg.hasPrev()) {
                findReg--;
                parsing_printf("\tChecking 0x%lx for register...\n", *findReg);
                if((*findReg.getInstruction().op_ptr()) == MOVREGMEM_REG 
                    || (*findReg.getInstruction().op_ptr()) == LEA) 
                {
                    // tableInsn now points to the instruction that loads
                    // what we presume to be the table base into
                    // /some register/

                    tableInsn = findReg.getInstruction();
                    tableInsnAddr = *findReg;
                    foundTableInsn = true;
                    parsing_printf("\tFound register at 0x%lx\n", *findReg);
                    break;    
                }
            }
        }

#if defined(cap_instruction_api)
   {
      /** 
       * AMD-64 gcc emits a re-accuring idiom of: 
       *         jmpq   *%r8
       *         movaps %xmm7,-0xf(%rax)
       *         movaps %xmm6,-0x1f(%rax)
       *         movaps %xmm5,-0x2f(%rax)
       *         movaps %xmm4,-0x3f(%rax)
       *         movaps %xmm3,-0x4f(%rax)
       *         movaps %xmm2,-0x5f(%rax)
       *         movaps %xmm1,-0x6f(%rax)
       *         movaps %xmm0,-0x7f(%rax)
       *         <other>
       *
       * The jump register is calculated in such a way that it'll be difficult
       * for our analysis to figure it out.  Instead we'll recognize the pattern
       * of the 'movaps'.  Note that the following instruction is also a valid jump
       * target
       **/
      std::set<Address> found;
      unsigned num_movaps_found = 0;
      InstructionDecoder d((const unsigned char *)img()->getPtrToInstruction(jumpAddr),
                           (img()->imageOffset() + img()->imageLength()) - jumpAddr);
      d.setMode(img()->getAddressWidth() == 8);
      Address cur = jumpAddr;
      unsigned last_insn_size = 0;
      InstructionAPI::Instruction i = d.decode();
      cur += i.size();
      for (;;) {
         InstructionAPI::Instruction insn = d.decode();
         //All insns in sequence are movaps
         if (insn.getOperation().getID() != e_movapd &&
             insn.getOperation().getID() != e_movaps) 
         {
            found.insert(cur);
            break;
         }
         //All insns are same size
         if (last_insn_size == 0)
            last_insn_size = insn.size();
         else if (last_insn_size != insn.size())
            break;

         num_movaps_found++;
         found.insert(cur);

         cur += insn.size();
      }
      if (num_movaps_found == 8) {
         //It's a match
         for (std::set<Address>::iterator i=found.begin(); i != found.end(); i++) {
            targets.insert(*i);
         }
         return true;
      }
   }
#endif

#if defined(cap_instruction_api)
		InstructionDecoder d((const unsigned char *)img()->getPtrToInstruction(tableInsnAddr), 
                             tableInsn.size());
   d.setMode(img()->getAddressWidth() == 8);
        Instruction tableInsn_iapi = d.decode();
        std::set<RegisterAST::Ptr> regsRead;
        tableInsn_iapi.getReadSet(regsRead);
#else
		std::set<RegisterAST::Ptr> regsRead;
#endif
        /* Previous comment said:	
        // search backward over the blocks that reach this one. we're looking
        // for a comparison on this register. if we find an assignment to the
        // register but don't find the comparison, we give up on this jump
        // table.

           XXX By "this register" it means "any register".

           This routine finds the comparison & conditional branch that guard
           the jump table.
        */
   bool foundMaxSwitch = findMaxSwitchInsn(currBlk, maxSwitch, branchInsn, foundJccAlongTaken, regsRead);

        // Searches for a thunk (call-thunk or rip-relative LEA as one
        // sees on x86-64). thunkOffset is filled in with the /base address
        // of the jump table/. Full stop.
        Address thunkOffset = 0;
        findThunkAndOffset(this, currBlk, thunkOffset);
	
        if( !foundMaxSwitch ) {
            parsing_printf("\tunable to fix max switch size\n");
            return false;
        }
 
            parsing_printf("\ttableInsn at 0x%lx\n",tableInsnAddr);
            if(thunkOffset) {
                parsing_printf("\tThunk-calculated table base address: 0x%lx\n",
                    thunkOffset);
            }

            // This routine does the work of computing target addresses
            // from the jump table. It:
            //   - computes the number of table entries from maxSwitch
            //   - figures out how to adjust that value based on the 
            //     branchInsn (e.g., jge vs jg)
            //   - computes the base address of the jump table from 
            //     tableInsn /IF/ isAddInJmp is est
            //   - or, uses thunkOffset as the base of the table if it
            //     is set (FIXME note that it could use both, adding
            //     thunkOffset to the base from tableInsn). This path
            //     of code is probably a bug but I don't know for sure.
            if( !ah.getMultipleJumpTargets( targets, tableInsn, 
                                   maxSwitch, branchInsn, isAddInJmp, 
                                   foundJccAlongTaken, thunkOffset))
            {
                return false;
            }
            else
            {
                return targets.size() > 0;
            }
}

bool image_func::archProcExceptionBlock(Address &catchStart, Address a)
{
   ExceptionBlock b;
  //    Symtab obj = img()->getObject();
    if (img()->getObject()->findCatchBlock(b,a)) {
        catchStart = b.catchStart();
        return true;
    } else {
        return false;
    }
}

bool image_func::archIsATailCall(InstrucIter &ah, 
                                 pdvector< instruction >& allInstructions)
{
  if(allInstructions.size() < 2) {
    return false;
  }
  
    unsigned numInsns = allInstructions.size() - 2;

    Address target = ah.getBranchTargetAddress();

    // Several cases
    //
    // If we are a direct jump AND destination is a known entry point OR an indirect jump
    //    if we have saved the FP && previous insn is POP_EBP, true
    //    if previous insn is LEAVE, true

    if( img()->findFuncByEntry( target ) ||
        ( *allInstructions[ numInsns ].ptr() == POP_EBP ||
        allInstructions[ numInsns ].isLeave() ))
    {
        return true;
    }
    else
        return false;
}

bool image_func::archIsIndirectTailCall(InstrucIter &ah)
{
  if(!ah.peekPrev()) return false;
  InstrucIter temp(ah);
  temp--;
  

  if (*(temp.getInstruction().op_ptr()) == POP_EBX) {
      //fprintf(stderr, "0x%lx: Indirect branch preceded by POP_EBX\n",
      //*temp);

      // No longer return true; not sure why it ever was. 
      return false;
  }
  if (*(temp.getInstruction().op_ptr()) == POP_EBP) {
      if (savesFramePointer()) {
          //fprintf(stderr, "0x%lx: Indirect branch preceded by POP_EBP, FRAMED\n",
          //*temp);
          return true;
      }
      else {
          //fprintf(stderr, "0x%lx: Indirect branch preceded by POP_EBP, FRAMELESS\n",
          //*temp);
          // Assume that EBP could hold real data, return false
          return false;
      }
  }
  if (temp.getInstruction().isLeave()) {
      //fprintf(stderr, "0x%lx: Indirect branch preceded by LEAVE\n",
      //*temp);
      return true;
  }

  return false;
}

bool image_func::archIsAbortOrInvalid(InstrucIter &ah)
{
    return ah.isAnAbortInstruction();
}

bool image_func::writesFPRs(unsigned level) {
    // Oh, we should be parsed by now...
    if (!parsed_) image_->analyzeIfNeeded();

    if (containsFPRWrites_ == unknown) {
        // Iterate down and find out...
        // We know if we have callees because we can
        // check the instPoints; no reason to iterate over.
        // We also cache callee values here for speed.

        if (level >= 3) return true; // Arbitrarily decided level 3 iteration.
        
        for (unsigned i = 0; i < calls.size(); i++) {
            if (calls[i]->func()) {
                if (calls[i]->func()->writesFPRs(level+1)) {
                    // One of our kids does... if we're top-level, cache it; in 
                    // any case, return
                    if (level == 0)
                        containsFPRWrites_ = used;
                    return true;
                }
            }
            else {
                // Indirect call... oh, yeah. 
                if (level == 0)
                    containsFPRWrites_ = used;
                return true;
            }
        }

        // No kids contain writes. See if our code does.
        InstrucIter ah(this);
        while (ah.hasMore()) {
            if (ah.isFPWrite()) {
                containsFPRWrites_ = used;
                return true;
            }
            ah++;
        }

        // No kids do, and we don't. Impressive.
        containsFPRWrites_ = unused;
        return false;
    }
    else if (containsFPRWrites_ == used)
        return true;
    else if (containsFPRWrites_ == unused)
        return false;

    fprintf(stderr, "ERROR: function %s, containsFPRWrites_ is %d (illegal value!)\n", 
	    symTabName().c_str(), containsFPRWrites_);
    
    assert(0);
    return false;
}
